# Copyright 2010 University of Chicago
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from contextlib import contextmanager
from UserDict import DictMixin

from pycassa.cassandra.ttypes import NotFoundException

__author__ = 'trhowe'

DEFAULT_COLUMN_COUNT = 1000

@contextmanager
def updating_node(node):
    yield node
    node.commit()


class Relationship(object):
    """
    Represents a directed connection between two nodes
    """

    def __init__(self, rel_id, source_node, target_node, data_store, type, args=dict()):
        """
        Create relationship.
        """
        self._rel_id = rel_id
        self.data_store = data_store
        self._type = type
        self.old_values = args
        self.new_values = {}
        self.new_values.update(self.old_values)
        self.relationship_factories = {}
        self._source_node = source_node
        self._target_node = target_node
        self.dirty = False

    @property
    def key(self):
        return self.old_values['rel_key']

    @property
    def rel_key(self):
        return self._rel_id

    @property
    def type(self):
        return self._type

    @property
    def source_node(self):
        return self._source_node

    @property
    def target_node(self):
        return self._target_node

    @property
    def attributes(self):
        for key in self.new_values:
            self.old_values[key] = self.new_values[key]

        class RelationshipFilteredDict(DictMixin):
            def __init__(self, delegate):
                self._delegate = delegate

            def _is_node_key(self, item):
                return item.startswith('target__') or item.startswith('source__')

            def _is_rel_key(self, item):
                return item.startswith('rel_key') or item.startswith('rel_type')

            def __getitem__(self, item):
                if  self._is_node_key(item) or self._is_rel_key(item):
                    raise KeyError
                else:
                    return self._delegate[item]

            def keys(self):
                return [
                    key for key in self._delegate.keys()
                    if not self._is_node_key(key) and not self._is_rel_key(key)
                ]

        return RelationshipFilteredDict(self.old_values)

    def __getitem__(self, item):
        if item in self.new_values:
            return self.new_values[item]
        else:
            return self.old_values[item]

    def __contains__(self, item):
        return item in self.new_values.keys() or item in self.old_values.keys()

    def __setitem__(self, key, value):
        self.new_values[key] = value
        self.dirty = True

    def __delitem__(self, key):
        if key in self.new_values.keys():
            del(self.new_values[key])
        
    def delete(self):
        self.data_store.delete_relationship(self._type, self._rel_id, self.source_node.type, self.source_node.key,
                                            self.target_node.type, self.target_node.key)

    def commit(self):
        self.data_store.create_relationship(self.type, self.source_node, self.target_node,
                                            key=self.key, args=self.new_values)

    def clear(self):
        self.new_values = {}
        self.new_values.update(self.old_values)
        self.dirty = False
    def __str__(self):
        return '%s: %s:%s -> %s:%s' % (
            self._type, self.source_node.type, self.source_node.key, self.target_node.type, self.target_node.key)

    def __eq__(self, other):
        if not isinstance(other, Relationship):
            return False
        return other.rel_key == self.rel_key and other.type == other.type

    def __cmp__(self, other):
        if not isinstance(other, Relationship):
            return -1
        return other.rel_key == self.rel_key and other.type == other.type


class RelationshipList(object):
    def __init__(self, relationships):
        self._relationships = relationships

    @property
    def single(self):
        if len(self._relationships) > 0:
            return self._relationships[0]
        else:
            return None

    def __len__(self):
        return len(self._relationships)

    def __iter__(self):
        for rel in self._relationships:
            yield rel


class RelationshipFactory(object):
    def __init__(self, data_store, parent_node, rel_type):
        self._rel_type = rel_type
        self._data_store = data_store
        self._parent_node = parent_node

    #TODO: specify order as from key vs timestamp
    def __call__(self, node, key=None, attributes=dict(), **kwargs):
        complete_attributes = {}
        complete_attributes.update(attributes)
        complete_attributes.update(kwargs)
        return self._data_store.create_relationship(self._rel_type, self._parent_node, node, key, complete_attributes)

    #TODO: Implement indexing solution here
    def __getitem__(self, item):
        try:
            relationships = []
            for relationship in self:
                if relationship.to_node.key == item:
                    relationships.append(relationship)
            return relationships
        except NotFoundException:
            return []

    def __contains__(self, item):
        return len(self.relationships_with(item)) > 0

    def relationships_with(self, node_key):
        return self._data_store.has_relationship(self._parent_node, node_key, self._rel_type)

    def get_outgoing(self, count):
        try:
            rels = self._data_store.get_outgoing_relationships(self._parent_node, self._rel_type, count=count)
        except NotFoundException:
            rels = []
        return RelationshipList(rels)

    def get_incoming(self, count):
        try:
            rels = self._data_store.get_incoming_relationships(self._parent_node, self._rel_type, count=count)
        except NotFoundException:
            rels = []
        return RelationshipList(rels)

    @property
    def parent_node(self):
        return self._parent_node

    @property
    def outgoing(self):
        try:
            rels = self._data_store.get_outgoing_relationships(self._parent_node, self._rel_type)
        except NotFoundException:
            rels = []
        return RelationshipList(rels)

    @property
    def incoming(self):
        try:
            rels = self._data_store.get_incoming_relationships(self._parent_node, self._rel_type)
        except NotFoundException:
            rels = []
        return RelationshipList(rels)


    def __len__(self):
        return len(self.outgoing) + len(self.incoming)

    def __iter__(self):
        for rel in self.outgoing:
            yield rel
        for rel in self.incoming:
            yield rel


class Node(object):
    def __init__(self, data_store, type, key, args=None):
        self._key = key
        self._data_store = data_store
        self._type = type
        self.old_values = args
        self.new_values = {}
        self.new_values.update(self.old_values)
        self.relationship_factories = {}
        self.dirty = False
        self._delete = False

    @property
    def key(self):
        return self._key

    @property
    def type(self):
        return self._type

    @property
    def relationships(self):
        class RelationshipsHolder(object):
            def __init__(self, data_store, node):
                self._data_store = data_store
                self._node = node

            @property
            def outgoing(self):
                return self._data_store.get_all_outgoing_relationships(self._node, DEFAULT_COLUMN_COUNT)

            @property
            def incoming(self):
                return self._data_store.get_all_incoming_relationships(self._node, DEFAULT_COLUMN_COUNT)

            def __len__(self):
                return len(self.outgoing) + len(self.incoming)

            def __iter__(self):
                rels = []
                rels.extend(self.outgoing)
                rels.extend(self.incoming)
                for rel in rels:
                    yield rel

        return RelationshipsHolder(self._data_store, self)


    def __getattr__(self, item):
        if hasattr(self.__dict__, item):
            return self.__dict__[item]
        else:
            relationship_factory = RelationshipFactory(self._data_store, self, item)
            self.relationship_factories[item] = relationship_factory
            return relationship_factory

    def __contains__(self, item):
        return item in self.attributes

    def __getitem__(self, item):
        if item in self.new_values:
            return self.new_values[item]

    def __setitem__(self, key, value):
        self.new_values[key] = value
        self.dirty = True

    def __delitem__(self, key):
        if key in self.new_values.keys():
            del(self.new_values[key])
            self.dirty = True

    @property
    def attributes(self):
        return self.new_values

    def delete(self):
        self._data_store.delete_node(self)
        self.node = None

    def commit(self):
        self._data_store.save_node(self)
        self.old_values = {}
        self.old_values.update(self.new_values)

    def clear(self):
        self.new_values = {}
        self.new_values.update(self.old_values)

    def __str__(self):
        return 'Node: %s => %s' % (self.type, self.key)

    def __cmp__(self, other):
        if not isinstance(other, Node):
            return False
        return other.type == self.type and other.key == self.key

    def __eq__(self, other):
        if not isinstance(other, Node):
            return False
        return other.type == self.type and other.key == self.key





from contextlib import contextmanager
import json
import string
import uuid
import datetime
from dateutil.parser import parse as date_parse
from pycassa.cassandra.ttypes import NotFoundException
from agamemnon.graph_constants import RELATIONSHIP_KEY_PATTERN, OUTBOUND_RELATIONSHIP_CF, RELATIONSHIP_INDEX, ENDPOINT_NAME_TEMPLATE, INBOUND_RELATIONSHIP_CF
import pycassa
from agamemnon.cassandra import CassandraDataStore
from agamemnon.memory import InMemoryDataStore
from agamemnon.exceptions import NodeNotFoundException
import agamemnon.primitives as prim

import logging

log = logging.getLogger(__name__)

class DataStore(object):
    def __init__(self, delegate):
        self.delegate = delegate

    @contextmanager
    def batch(self):
        self.delegate.start_batch()
        yield
        self.delegate.commit_batch()

    def get(self, type, row_key, column_start=None, columns=None, column_finish=None, column_count=100,
            super_column_key=None):
        column_family = self.delegate.get_cf(type)
        args = {}
        if columns is not None:
            args['columns'] = columns
        if super_column_key is not None:
            args['super_column'] = super_column_key
        if column_start is not None:
            args['column_start'] = column_start
        if column_finish is not None:
            args['column_finish'] = column_finish
        args['column_count'] = column_count


        return self.deserialize_value(column_family.get(row_key, **args))

    def delete(self, type, key, super_key=None, columns=None):
        if super_key is None:
            self.delegate.remove(self.get_cf(type), key, columns=columns)
        else:
            self.delegate.remove(self.get_cf(type), key, super_column=super_key, columns=columns)

    def insert(self, type, key, args, super_key=None):
        if not self.delegate.cf_exists(type):
            column_family = self.delegate.create_cf(type)
        else:
            column_family = self.delegate.get_cf(type)
        serialized = self.serialize_columns(args)
        if super_key is None:
            column_family.insert(key, serialized)
        else:
            column_family.insert(key, {super_key: serialized})

    def get_all_outgoing_relationships(self, source_node, count=100):
        source_key = RELATIONSHIP_KEY_PATTERN % (source_node.type, source_node.key)
        try:
            super_columns = self.get(OUTBOUND_RELATIONSHIP_CF, source_key, column_count=count)
        except NotFoundException:
            super_columns = {}
        return [self.get_outgoing_relationship(super_column[1]['rel_type'], source_node, super_column) for super_column in
                super_columns.items() if len(super_column[1]) > 0]

    def get_all_incoming_relationships(self, target_node, count=100):
        source_key = RELATIONSHIP_KEY_PATTERN % (target_node.type, target_node.key)
        try:
            super_columns = self.get(INBOUND_RELATIONSHIP_CF, source_key, column_count=count)
        except NotFoundException:
            super_columns = {}
        return [self.get_incoming_relationship(super_column[1]['rel_type'], target_node, super_column) for super_column in
                super_columns.items() if len(super_column[1]) > 0]

    def get_outgoing_relationships(self, source_node, rel_type, count=100):
        source_key = RELATIONSHIP_KEY_PATTERN % (source_node.type, source_node.key)
        #Ok, this is weird.  So in order to get a column slice, you need to provide a start that is <= your first column
        #id, and a finish which is >= your last column.  Since our columns are sorted by ascii, this means we need to go
        #from rel_type_ to rel_type` because "`" is the char 1 greater than "_", so this will get anything which starts
        #rel_type_.  Now I realize that this could problems when there is a "_" in the relationship name, so we will
        #probably need a different delimiter.
        #TODO: fix delimiter
        try:
            super_columns = self.get(OUTBOUND_RELATIONSHIP_CF, source_key, column_start='%s__' % rel_type,
                                     column_finish='%s_`' % rel_type,
                                     column_count=count)
        except NotFoundException:
            super_columns = {}
        return [self.get_outgoing_relationship(rel_type, source_node, super_column) for super_column in
                super_columns.items()]


    def get_incoming_relationships(self, target_node, rel_type, count=100):
        target_key = RELATIONSHIP_KEY_PATTERN % (target_node.type, target_node.key)
        #Ok, this is weird.  So in order to get a column slice, you need to provide a start that is <= your first column
        #id, and a finish which is >= your last column.  Since our columns are sorted by ascii, this means we need to go
        #from rel_type_ to rel_type` because "`" is the char 1 greater than "_", so this will get anything which starts
        #rel_type_.  Now I realize that this could problems when there is a "_" in the relationship name, so we will
        #probably need a different delimiter.
        #TODO: fix delimiter
        try:
            super_columns = self.get(INBOUND_RELATIONSHIP_CF, target_key, column_start='%s__' % rel_type,
                                     column_finish='%s_`' % rel_type,
                                     column_count=count)
        except NotFoundException:
            super_columns = {}
        return [self.get_incoming_relationship(rel_type, target_node, super_column) for super_column in
                super_columns.items()]

    def get_outgoing_relationship(self, rel_type, source_node, super_column):
        """
        Process the contents of a SuperColumn to extract the relationship and to_node properties and return
        a constructed relationship
        """
        rel_key = super_column[0]
        target_node_key = None
        target_node_type = None
        target_attributes = {}
        rel_attributes = {}
        for column in super_column[1].keys():
            value = super_column[1][column]
            if column == 'target__type':
                target_node_type = value
            elif column == 'target__key':
                target_node_key = value
            elif column.startswith('target__'):
                target_attributes[column[8:]] = value
            else:
                rel_attributes[column] = value
        return prim.Relationship(rel_key, source_node,
                                 prim.Node(self, target_node_type, target_node_key, target_attributes), self
                                 , rel_type,
                                 rel_attributes)

    def get_incoming_relationship(self, rel_type, target_node, super_column):
        """
        Process the contents of a SuperColumn to extract an incoming relationship and the associated from_node and
        return a constructed relationship
        """
        rel_key = super_column[0]
        source_node_key = None
        source_node_type = None
        source_attributes = {}
        rel_attributes = {}
        for column in super_column[1].keys():
            value = super_column[1][column]
            if column == 'source__type':
                source_node_type = value
            elif column == 'source__key':
                source_node_key = value
            elif column.startswith('source__'):
                source_attributes[column[8:]] = value
            else:
                rel_attributes[column] = value
        return prim.Relationship(rel_key, prim.Node(self, source_node_type, source_node_key, source_attributes),
                                 target_node,
                                 self, rel_type, rel_attributes)


    def delete_relationship(self, rel_type, rel_id, from_type, from_key, to_type, to_key):
        rel_from_key = ENDPOINT_NAME_TEMPLATE % (from_type, from_key)
        rel_to_key = ENDPOINT_NAME_TEMPLATE % (to_type, to_key)

        with self.batch():
            self.delete(INBOUND_RELATIONSHIP_CF, rel_to_key, super_key=rel_id)
            self.delete(OUTBOUND_RELATIONSHIP_CF, rel_from_key, super_key=rel_id)
            self.delete(RELATIONSHIP_INDEX, rel_to_key, super_key=from_key, columns=[rel_type])
            self.delete(RELATIONSHIP_INDEX, rel_from_key, super_key=to_key, columns=[rel_type])

    def create_relationship(self, rel_type, source_node, target_node, key=None, args=dict()):
        if key is None:
            key = str(uuid.uuid4())
            #node relationship types
        rel_key = RELATIONSHIP_KEY_PATTERN % (rel_type, key)
        with self.batch():
            #outbound_cf
            columns = {'rel_type': rel_type, 'rel_key': key}
            #add relationship attributes
            columns.update(args)
            rel_attr = dict(columns)
            #add target attributes
            columns['target__type'] = target_node.type.encode('ascii')
            columns['target__key'] = target_node.key.encode('ascii')
            target_attributes = target_node.attributes
            for attribute_key in target_attributes.keys():
                columns['target__%s' % attribute_key] = target_attributes[attribute_key]
            columns['source__type'] = source_node.type.encode('ascii')
            columns['source__key'] = source_node.key.encode('ascii')
            source_attributes = source_node.attributes
            for attribute_key in source_attributes.keys():
                columns['source__%s' % attribute_key] = source_attributes[attribute_key]

            source_key = ENDPOINT_NAME_TEMPLATE % (source_node.type, source_node.key)
            target_key = ENDPOINT_NAME_TEMPLATE % (target_node.type, target_node.key)
            serialized = self.serialize_columns(columns)
            self.insert(OUTBOUND_RELATIONSHIP_CF, source_key, {rel_key: serialized})
            self.insert(INBOUND_RELATIONSHIP_CF, target_key, {rel_key: serialized})

            #            relationship_index_cf = self.delegate.get_cf(RELATIONSHIP_INDEX)
            # Add entries in the relationship index
            self.insert(RELATIONSHIP_INDEX, source_key, {target_node.key: {rel_type: '%s__outgoing' % rel_key}})
            self.insert(RELATIONSHIP_INDEX, target_key, {source_node.key: {rel_type: '%s__incoming' % rel_key}})

        #created relationship object
        return prim.Relationship(rel_key, source_node, target_node, self, rel_type, rel_attr)

    def has_relationship(self, node_a, node_b_key, rel_type):
        """
        This determines if two nodes have a relationship of the specified type.

        > ds = DataStore()
        > node_a =

        """
        index = self.delegate.get_cf(RELATIONSHIP_INDEX)
        node_a_row_key = ENDPOINT_NAME_TEMPLATE % (node_a.type, node_a.key)
        rel_list = []
        try:
            rels = self.get(RELATIONSHIP_INDEX, node_a_row_key, super_column_key=node_b_key, columns=[rel_type])
            for rel in rels.values():
                if rel.endswith('__incoming'):
                    rel_id = string.replace(rel, '__incoming', '')
                    super_column = self.get(INBOUND_RELATIONSHIP_CF, node_a_row_key, columns=[rel_id]).items()[0]
                    relationship = self.get_incoming_relationship(rel_type, node_a, super_column)
                elif rel.endswith('__outgoing'):
                    rel_id = string.replace(rel, '__outgoing', '')
                    super_column = self.get(OUTBOUND_RELATIONSHIP_CF, node_a_row_key, columns=[rel_id]).items()[0]
                    relationship = self.get_outgoing_relationship(rel_type, node_a, super_column)
                else:
                    continue

                rel_list.append(relationship)
        except NotFoundException:
            pass
        return rel_list

    def create_node(self, type, key, args=None, reference=False):
        if args is None:
            args = {}
        node = prim.Node(self, type, key, args)

        #since node won't get created without args, we will include __id by default
        args["__id"] = key
        serialized = self.serialize_columns(args)
        self.insert(type, key, serialized)
        if not reference:
            #this adds the created node to the reference node for this type of object
            #that reference node functions as an index to easily access all nodes of a specific type
            reference_node = self.get_reference_node(type)
            reference_node.instance(node, key=key)
        return node

    def delete_node(self, node):
        relationships = node.relationships
        with self.batch():
           for rel in relationships:
                rel.delete()
                self.delete(node.type, node.key)

    def save_node(self, node):
        """
        This needs to update the entry in the type table as well as all of the relationships
        """
        self.insert(node.type, node.key, self.serialize_columns(node.attributes))
        columns_to_remove = []
        for key in node.old_values:
            if not key in node.new_values:
                columns_to_remove.append(key)
        if len(columns_to_remove) > 0:
            self.remove(self.get_cf(node.type), node.key, columns=columns_to_remove)
        source_key = ENDPOINT_NAME_TEMPLATE % (node.type, node.key)
        target_key = ENDPOINT_NAME_TEMPLATE % (node.type, node.key)

        try:
            outbound_results = self.get(OUTBOUND_RELATIONSHIP_CF, source_key)
        except NotFoundException:
            outbound_results = {}
        try:
            inbound_results = self.get(INBOUND_RELATIONSHIP_CF, target_key)
        except NotFoundException:
            inbound_results = {}
        outbound_columns = {'source__type': node.type.encode('utf-8'), 'source__key': node.key.encode('utf-8')}
        node_attributes = node.attributes
        for attribute_key in node.attributes.keys():
            outbound_columns['source__%s' % attribute_key] = node_attributes[attribute_key]
        for key in outbound_results.keys():
            target = outbound_results[key]
            target_key = ENDPOINT_NAME_TEMPLATE % (target['target__type'], target['target__key'])
            target.update(outbound_columns)
            serialized = self.serialize_columns(target)
            self.insert(OUTBOUND_RELATIONSHIP_CF, source_key, serialized, key)
            self.insert(INBOUND_RELATIONSHIP_CF, target_key, serialized, key)
            if len(columns_to_remove):
                self.delete(OUTBOUND_RELATIONSHIP_CF, source_key, super_key=key,
                        columns=['source__%s' % column for column in columns_to_remove])
                self.delete(INBOUND_RELATIONSHIP_CF, target_key, super_key=key,
                        columns=['source__%s' % column for column in columns_to_remove])
        inbound_columns = {'target__type': node.type.encode('utf-8'), 'target__key': node.key.encode('utf-8')}
        for attribute_key in node.attributes.keys():
            inbound_columns['target__%s' % attribute_key] = node_attributes[attribute_key]
        for key in inbound_results.keys():
            source = inbound_results[key]
            source_key = ENDPOINT_NAME_TEMPLATE % (source['source__type'], source['source__key'])
            target_key = ENDPOINT_NAME_TEMPLATE % (node.type, node.key)
            source.update(inbound_columns)
            serialized = self.serialize_columns(source)
            self.insert(OUTBOUND_RELATIONSHIP_CF, source_key, serialized, key)
            self.insert(INBOUND_RELATIONSHIP_CF, target_key, serialized, key)
            if len(columns_to_remove):
                self.delete(OUTBOUND_RELATIONSHIP_CF, source_key, super_key=key,
                        columns=['target__%s' % column for column in columns_to_remove])
                self.delete(INBOUND_RELATIONSHIP_CF, target_key, super_key=key,
                        columns=['target__%s' % column for column in columns_to_remove])



    def get_node(self, type, key):
        try:
            values = self.get(type, key)
        except NotFoundException:
            raise NodeNotFoundException()
        return prim.Node(self, type, key, values)

    def get_reference_node(self, name='reference'):
#        """
#        Nodes returned here are very easily referenced by name and then function as an index for all attached nodes
#        The most typical use case is to index all of the nodes of a certain type, but the functionality is not limited
#        to this.
#        """
#        try:
#            node = self.get_node('reference', name)
#        except NodeNotFoundException:
#            node = self.create_node('reference', name, {'reference': 'reference'}, reference=True)
#            self.get_reference_node().instance(node)
#        return node
        try:
            ref_node = self.get_node('reference', name)
        except NodeNotFoundException:
            ref_node = self.create_node('reference', name, {'reference': 'reference'}, reference=True)
            self.get_reference_node().instance(ref_node, key='reference_%s' % name)

        return ref_node

    def deserialize_value(self, value):
        if isinstance(value, dict):
            return self.deserialize_columns(value)
        if not value.startswith('$'):
            return value
        type = value[:2]
        content = value[2:]
        if type == '$b':
            if content == 'True':
                return True
            if content == 'False':
                return False
        elif type == '$i':
            return int(content)
        elif type == '$l':
            return long(content)
        elif type == '$f':
            return float(content)
        elif type == '$t':
            return date_parse(content)

    def serialize_value(self, value):
        if isinstance(value, bool):
            return '$b%r' % value
        elif isinstance(value, int):
            return '$i%r' % value
        elif isinstance(value, long):
            return '$l%r' % value
        elif isinstance(value, float):
            return '$f%r' % value
        elif isinstance(value, str):
            return value
        elif isinstance(value, unicode):
            return value.encode('utf-8')
        elif isinstance(value, dict):
            return self.serialize_columns(value)
        elif isinstance(value, datetime.datetime):
            return '$t%s' % value.isoformat()
        else:
            raise TypeError('Cannot serialize: %s' % type(value))

    def deserialize_columns(self, columns):
        return dict([(key, self.deserialize_value(value))
        for key, value in columns.items()
        if value is not None])

    def serialize_columns(self, columns):
        return dict([(key, self.serialize_value(value))
        for key, value in columns.items()
        if value is not None])

    def __getattr__(self, item):
        if not item in self.__dict__:
            return getattr(self.delegate, item)


def load_from_settings(settings, prefix='agamemnon.'):
    if settings["%skeyspace" % prefix] == 'memory':
        ds_to_wrap = InMemoryDataStore()
    else:
        ds_to_wrap = CassandraDataStore(settings['%skeyspace' % prefix],
                                        pycassa.connect(settings["%skeyspace" % prefix],
                                                        json.loads(settings["%shost_list" % prefix])),
                                        system_manager=pycassa.system_manager.SystemManager(
                                            json.loads(settings["%shost_list" % prefix])[0]))
    return DataStore(ds_to_wrap)



# -*- encoding: ISO-8859-5 -*-
import random
from unittest import TestCase
from agamemnon import cassandra
from agamemnon.factory import load_from_settings
from agamemnon.primitives import updating_node

in_memory = False

class AgamemnonTests(TestCase):

    def set_up_cassandra(self):
        host_list = '["localhost:9160"]'
        keyspace = 'agamemnontests'
        try:
            cassandra.drop_keyspace(host_list, keyspace)
        except Exception:
            pass
        cassandra.create_keyspace(host_list, keyspace)
        self.ds = load_from_settings({
            'agamemnon.host_list': host_list,
            "agamemnon.keyspace": keyspace
        })

    def set_up_in_memory(self):
        self.ds = load_from_settings({'agamemnon.keyspace': 'memory'})
    
    def create_node(self, node_type, id):
        attributes = {
            'boolean': True,
            'integer': id,
            'long': long(1000),
            'float': 1.5434235,
            'string': 'name%s' % id,
            'unicode': 'пярст'
        }
        key = 'node_%s' % id
        self.ds.create_node(node_type, key, attributes)
        node = self.ds.get_node(node_type, key)
        self.failUnlessEqual(key, node.key)
        self.failUnlessEqual(node_type, node.type)
        return key, attributes

    def containment(self, node_type, node):
        reference_node = self.ds.get_reference_node(node_type)
        test_reference_nodes = [rel.source_node for rel in node.instance.incoming]
        self.failUnlessEqual(1, len(test_reference_nodes))
        self.failUnlessEqual(reference_node, test_reference_nodes[0])
        ref_ref_node = self.ds.get_reference_node()
        test_reference_nodes = [rel.target_node for rel in ref_ref_node.instance.outgoing]
        self.failUnlessEqual(2, len(test_reference_nodes))
        self.failUnlessEqual(sorted([ref_ref_node, reference_node]), sorted(test_reference_nodes))
        self.failUnless(node_type in ref_ref_node.instance)
        self.failUnless(ref_ref_node.key in reference_node.instance)

    def get_set_attributes(self, node, attributes):
        self.failUnlessEqual(attributes, node.attributes)
        node['new_attribute'] = 'sample attr'
        node.commit()
        node = self.ds.get_node(node.type, node.key)
        self.failUnlessEqual('sample attr', node['new_attribute'])
        self.failIfEqual(attributes, node.attributes)
        # Test the context manager
        node = self.ds.get_node(node.type, node.key)
        with updating_node(node):
            node['new_attribute'] = 'new sample attr'
        node = self.ds.get_node(node.type, node.key)
        self.failUnlessEqual('new sample attr', node['new_attribute'])
        self.failIfEqual(attributes, node.attributes)
        with updating_node(node):
            del(node['new_attribute'])
        node = self.ds.get_node(node.type, node.key)
        if 'new_attribute' in node:
            print "We should not have found 'new_attribute' = %s" % node['new_attribute']
            self.fail()


    def create_random_relationship(self, node, target_type, node_list):
        target_key, target_attributes = random.choice(node_list)
        while target_key == node.key and not target_key in node.is_related_to:
            target_key, target_attributes = random.choice(node_list)
        attributes = {
            'int': 10,
            'float': 2.3,
            'long': long(10),
            'boolean': True,
            'string': 'string',
            'unicode': 'пярст'
        }
        kw_args = {
            'test_kwarg': 'kw'
        }
        target_node = self.ds.get_node(target_type, target_key)
        rel = node.is_related_to(target_node, attributes=attributes, **kw_args)
        self.failUnless(target_key in node.is_related_to)
        self.failUnless(node.key in target_node.is_related_to)
        rel_to_target = target_node.is_related_to.relationships_with(node.key)[0]
        self.failUnlessEqual(rel, rel_to_target)
        complete_attributes = {}
        complete_attributes.update(attributes)
        complete_attributes.update(kw_args)
        test_attributes = rel_to_target.attributes
        for key in complete_attributes.keys():
            self.failUnlessEqual(complete_attributes[key], test_attributes[key])
        self.failUnlessEqual(len(complete_attributes), len(test_attributes))
        self.failUnlessEqual(rel.key, rel_to_target.key)
        in_outbound_relationships = False
        for rel in node.is_related_to.outgoing:
            if rel.target_node.key == target_key:
                in_outbound_relationships = True
        self.failUnless(in_outbound_relationships)
        in_inbound_relationships = False
        for rel in target_node.is_related_to.incoming:
            if rel.source_node.key == node.key:
                in_inbound_relationships = True
        self.failUnless(in_inbound_relationships)
        rel['dummy_variable'] = 'dummy'
        rel_attributes = rel.attributes
        self.failIfEqual(attributes, rel.attributes)
        self.failUnlessEqual('dummy', rel_attributes['dummy_variable'])
        del(rel['dummy_variable'])
        try:
            rel['dummy_variable']
        except KeyError:
            pass

        rel['int'] = 20
        rel.commit()
        rel_to_target = target_node.is_related_to.relationships_with(node.key)[0]
        if rel_to_target.key == rel.key:
            self.failUnlessEqual(20, rel_to_target['int'])
        return node, target_node

    def delete_relationships(self, source, target):
        source_initial_rel_count = len(source.relationships)
        target_initial_rel_count = len(target.relationships)
        self.failUnless(target.key in source.is_related_to)
        self.failUnless(source.key in target.is_related_to)
        rel_list = source.is_related_to.relationships_with(target.key)
        self.failUnlessEqual(1, len(rel_list))
        rel = rel_list[0]
        rel.delete()
        self.failIf(target.key in source.is_related_to)
        self.failIf(source.key in target.is_related_to)
        source_post_delete_count = len(source.relationships)
        target_post_delete_count = len(target.relationships)
        self.failUnlessEqual(source_initial_rel_count - 1, source_post_delete_count)
        self.failUnlessEqual(target_initial_rel_count - 1, target_post_delete_count)
        return rel

    def test_cassandra(self):
        """
        Test the cassandra data store
        """
        self.set_up_cassandra()
        self.one_node_type_one_relationship_type()

    def test_in_memory(self):
        """
        Test the in_memory data store
        """
        self.set_up_in_memory()
        self.one_node_type_one_relationship_type()

    def one_node_type_one_relationship_type(self):
        """
        Tests for one node type and one relationship type.
        """
        node_type = "type_a"

        node_list = []
        for i in xrange(100):
            node_list.append(self.create_node(node_type, i))
        for key, attributes in node_list:
            node = self.ds.get_node(node_type, key)
            # test the basic details of the reference node including containment
            self.containment(node_type, node)
            # test updating the attributes of the node
            self.get_set_attributes(node, attributes)
        #Generate "random" network
        relationships = []
        for key, attributes in node_list:
            node = self.ds.get_node(node_type, key)
            for i in range(5):
                relationships.append(self.create_random_relationship(node, node_type, node_list))

        random_relationships = []
        for i in xrange(10):
            source, target = random.choice(relationships)
            self.delete_relationships(source, target)
            relationships.remove((source, target))

        for source, target in random_relationships: self.delete_relationships(source, target)

        #delete node
        deleted_nodes = []
        for i in xrange(10):
            source, target = random.choice(relationships)
            deleted_nodes.append(source)
            relationships_to_delete = [rel for rel in source.relationships]
            source.delete()
            for deleted_rel in relationships_to_delete:
                target_incoming_relationships = [rel for rel in target.is_related_to.incoming]
                self.failIf(source.key in target.is_related_to)
                self.failIf(deleted_rel in target_incoming_relationships)



import operator
from ordereddict import OrderedDict
from pycassa.cassandra.ttypes import NotFoundException
from agamemnon.graph_constants import ASCII
import logging

log = logging.getLogger()

class InMemoryDataStore(object):
    def __init__(self):
        self.tables = {}
        self.transactions = {}
        self.batch_count = 0
        self.in_batch = False

    def get_cf(self, cf_name):
        if not cf_name in self.tables:
            self.tables[cf_name] = self.create_cf(cf_name)
        return self.tables[cf_name]

    def create_cf(self, type, column_type=ASCII, super=False, index_columns=list()):
        self.tables[type] = ColumnFamily(type, column_type)
        return self.tables[type]

    def cf_exists(self, type):
        return type in self.tables.keys()

    def insert(self, cf, row, columns):
        def execute():
            cf.insert(row, columns)

        if self.in_batch:
            self.transactions[('insert', cf.name, row, tuple(columns.keys()))] = execute
        else:
            execute()

    def remove(self, cf, row, columns=None, super_column=None):
        def execute():
            cf.remove(row, columns=columns, super_column=super_column)
            
        if self.in_batch:
            if columns is not None and super_column is not None\
            and not ('remove', cf.name, row, super_column) in self.transactions.keys():
                key = ('remove', cf.name, row, tuple(columns), super_column)
            elif columns is not None and not ('remove', cf.name, row) in self.transactions.keys():
                key = ('remove', cf.name, row, tuple(columns))
            elif super_column is not None:
                key = ('remove', cf.name, row, super_column)
            else:
                key = ('remove', cf.name, row)
            if not key in self.transactions.keys() and not ('remove', cf.name, row) in self.transactions.keys():
                self.transactions[key] = execute
        else:
            execute()

    def start_batch(self):
        self.in_batch = True
        self.batch_count += 1

    def commit_batch(self):
        self.batch_count -= 1
        if not self.batch_count:
            for key, value in self.transactions.items():
                value()
            self.transactions.clear()
            self.in_batch = False


class ColumnFamily(object):
    def __init__(self, name, sort, super=False):
        self.data = OrderedDict()
        self.sort = sort
        self.name = name
        self.super = super

    def get(self, row, columns=None, column_start=None, super_column=None, column_finish=None, column_count=100):
        try:
            if columns is None and column_start is None and super_column is None:
                results = self.data[row]
            else:
                if super_column is None:
                    data_columns = self.data[row]
                else:
                    data_columns = self.data[row][super_column]
                results = {}
                count = 0
                if columns is not None:
                    for c in columns:
                        results[c] = data_columns[c]
                else:
                    for c in data_columns.keys():
                        if count > column_count:
                            break
                        if column_start is not None and column_finish is not None:
                            if ((cmp(c, column_start) > 0
                                and cmp(c, column_finish) < 0)
                                or cmp(c, column_finish) == 0
                                or cmp(c, column_start) == 0):

                                results[c] = data_columns[c]
                                count += 1
                        else:
                            results[c] = data_columns[c]
                            count += 1
            if not len(results):
                raise NotFoundException
            for key, value in results.items():
                if isinstance(value, dict) and len(value) == 0:
                    del(results[key])
                if value is None:
                    del(results[key])
            return results
        except KeyError:
            raise NotFoundException

    def insert(self, row, columns, ttl=None):
        if not row in self.data:
            self.data[row] = OrderedDict()
        if self.sort == ASCII:
            sorted_columns = sorted(columns.iteritems(), key=operator.itemgetter(0))
            for column in sorted_columns:
                self.data[row][column[0]] = column[1]

                #        if ttl is not None:
                #            def delete():
                #                for c in columns.keys():
                #                    del(self.data[row][c])
                #            Timer(ttl, delete, ()).start()

    def remove(self, row, columns=None, super_column=None):
        try:
            if columns is None and super_column is None:
                row_data = self.data[row]
                for key, value in row_data.items():
                    if isinstance(value, OrderedDict):
                        value.clear()
                    else:
                        del(row_data[key])
            elif columns is not None and super_column is None:
                row_data = self.data[row]
                for c in columns:
                    if c in row_data:
                        value = row_data[c]
                        if isinstance(value, OrderedDict):
                            value.clear()
                        else:
                            del(row_data[c])
            elif columns is None and super_column is not None:
                for key, value in self.data[row][super_column].items():
                    if isinstance(value, dict):
                        value.clear()
                    else:
                        del(self.data[row][super_column][key])
            else:
                sc = self.data[row][super_column]
                for c in columns:
                    if c in sc:
                        del(sc[c])
        except KeyError:
            raise NotFoundException

    def get_indexed_slices(self, index_clause):
        expression = index_clause.expressions[0]
        column_name = expression.column_name
        value = expression.value
        for i in self.data.items():
            if i[1][column_name] == value:
                yield i[0], i[1]
  



# Copyright 2010 University of Chicago
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from agamemnon.factory import load_from_settings
from agamemnon.exceptions import NoTransactionError, NodeNotFoundException

log = logging.getLogger(__name__)

def DFS(node, relationship_type, return_predicate=None):
    visited = set([node.key])
    S = [relationship for relationship in getattr(node, relationship_type)]
    while S:
        p = S.pop()
        relationship = p
        child = relationship.target_node
        if child.key not in visited:
            if return_predicate is not None and return_predicate(relationship, child):
                visited.add(child.key)
                yield child
            elif return_predicate is None:
                visited.add(child.key)
                yield child
            if hasattr(child, relationship_type):
                visited.add(child.key)
                S.extend([relationship for relationship in getattr(child, relationship_type)])




import unittest
import time

from tempfile import mkdtemp

from rdflib.term import URIRef, BNode, Literal
from rdflib.namespace import RDF, Namespace
from rdflib.graph import Graph

import rdflib.plugin

from agamemnon.rdf_store import AgamemnonStore

import uuid

import logging
log = logging.getLogger(__name__)

class GraphMemoryTestCase(unittest.TestCase):
    store_name = 'Agamemnon'
    settings1 = {
        'agamemnon.keyspace' : 'memory',
        'agamemnon.host_list' : '["localhost:9160"]',
        'agamemnon.rdf_namespace_base' : 'http://www.example.org/rdf/',
    }
    settings2 = {
        'agamemnon.keyspace' : 'memory',
        'agamemnon.host_list' : '["localhost:9160"]',
        'agamemnon.rdf_namespace_base' : 'http://www.example.org/rdf/',
    }

    def setUp(self):
        self.graph1 = Graph(store=self.store_name)
        self.graph2 = Graph(store=self.store_name)

        self.graph1.open(self.settings1, True)
        self.graph2.open(self.settings2, True)

        self.oNS = Namespace("http://www.example.org/rdf/things#")
        self.sNS = Namespace("http://www.example.org/rdf/people#")
        self.pNS = Namespace("http://www.example.org/rdf/relations/")

        self.graph1.bind('people',self.sNS)
        self.graph1.bind('relations',self.pNS)
        self.graph1.bind('things',self.oNS)
        self.graph2.bind('people',self.sNS)
        self.graph2.bind('relations',self.pNS)
        self.graph2.bind('things',self.oNS)

        self.michel = self.sNS.michel
        self.tarek = self.sNS.tarek
        self.alice = self.sNS.alice
        self.bob = self.sNS.bob
        self.likes = self.pNS.likes
        self.hates = self.pNS.hates
        self.named = self.pNS.named
        self.pizza = self.oNS.pizza
        self.cheese = self.oNS.cheese

    def tearDown(self):
        self.graph1.close()
        self.graph2.close()

    def addStuff(self,graph):
        graph.add((self.tarek, self.likes, self.pizza))
        graph.add((self.tarek, self.likes, self.cheese))
        graph.add((self.michel, self.likes, self.pizza))
        graph.add((self.michel, self.likes, self.cheese))
        graph.add((self.bob, self.likes, self.cheese))
        graph.add((self.bob, self.hates, self.pizza))
        graph.add((self.bob, self.hates, self.michel)) # gasp!
        graph.add((self.bob, self.named, Literal("Bob")))

    def removeStuff(self,graph):
        graph.remove((None, None, None))


    def testBind(self):
        store = self.graph1.store
        self.assertEqual(store.namespace(""), Namespace("http://www.example.org/rdf/"))
        self.assertEqual(store.namespace('people'),self.sNS)
        self.assertEqual(store.namespace('relations'),self.pNS)
        self.assertEqual(store.namespace('things'),self.oNS)
        self.assertEqual(store.namespace('blech'),None)

        self.assertEqual("", store.prefix(Namespace("http://www.example.org/rdf/")))
        self.assertEqual('people',store.prefix(self.sNS))
        self.assertEqual('relations',store.prefix(self.pNS))
        self.assertEqual('things',store.prefix(self.oNS))
        self.assertEqual(None,store.prefix("blech"))

        self.assertEqual(len(list(self.graph1.namespaces())), 7)

    def testRelationshipToUri(self):
        uri = self.graph1.store.rel_type_to_ident('likes')
        self.assertEqual(uri, URIRef("http://www.example.org/rdf/likes"))

        uri = self.graph1.store.rel_type_to_ident('emotions:likes')
        self.assertEqual(uri, URIRef("emotions:likes"))

        self.graph1.bind('emotions','http://www.emo.org/')
        uri = self.graph1.store.rel_type_to_ident('emotions:likes')
        self.assertEqual(uri, URIRef("http://www.emo.org/likes"))

    def testNodeToUri(self):
        node = self.graph1.store._ds.create_node('blah', 'bleh')
        uri = self.graph1.store.node_to_ident(node)
        self.assertEqual(uri, URIRef("http://www.example.org/rdf/blah#bleh"))

        self.graph1.bind("bibble", "http://www.bibble.com/rdf/bibble#")
        node = self.graph1.store._ds.create_node('bibble', 'babble')
        uri = self.graph1.store.node_to_ident(node)
        self.assertEqual(uri, URIRef("http://www.bibble.com/rdf/bibble#babble"))

    def testUriToRelationship(self):
        rel_type = self.graph1.store.ident_to_rel_type(URIRef("http://www.example.org/rdf/likes"))
        self.assertEqual(rel_type, 'likes')

        rel_type = self.graph1.store.ident_to_rel_type(URIRef('emotions:likes'))
        prefix, rel_type = rel_type.split(":",1)
        uuid.UUID(prefix.replace("_","-"))
        self.assertEqual(rel_type, "likes")

        self.graph1.bind('emotions','http://www.emo.org/')
        rel_type = self.graph1.store.ident_to_rel_type(URIRef("http://www.emo.org/likes"))
        self.assertEqual(rel_type, 'emotions:likes')
        

    def testUriToNode(self):
        #test unbound uri
        uri = URIRef("http://www.example.org/rdf/blah#bleh")
        node = self.graph1.store.ident_to_node(uri, True)
        uuid.UUID(node.type.replace("_","-"))
        self.assertEqual(node.key, "bleh")

        #test unbound uri with trailing /
        uri = URIRef("http://www.example.org/blah/bleh/")
        node = self.graph1.store.ident_to_node(uri, True)
        uuid.UUID(node.type.replace("_","-"))
        self.assertEqual(node.key, "bleh/")

        # teset bound uri
        self.graph1.bind("bibble", "http://www.bibble.com/rdf/bibble#")
        uri = URIRef("http://www.bibble.com/rdf/bibble#babble")
        node = self.graph1.store.ident_to_node(uri, True)
        self.assertEqual(node.type, "bibble")
        self.assertEqual(node.key, "babble")

        # make sure if we reference a predicate as a subject or object, we will
        # still be able recover the correct uri
        uri = URIRef("http://www.example.org/rdf/doit")
        node = self.graph1.store.ident_to_node(uri, True)
        self.assertEqual(self.graph1.store.node_to_ident(node), uri)

    def testAdd(self):
        self.addStuff(self.graph1)

    def testRemove(self):
        self.addStuff(self.graph1)
        self.removeStuff(self.graph1)

    def testTriples(self):
        tarek = self.tarek
        michel = self.michel
        bob = self.bob
        likes = self.likes
        hates = self.hates
        pizza = self.pizza
        cheese = self.cheese
        asserte = self.assertEquals
        triples = self.graph1.triples
        named = self.named
        Any = None

        self.addStuff(self.graph1)

        # unbound subjects
        asserte(len(list(triples((Any, likes, pizza)))), 2)
        asserte(len(list(triples((Any, hates, pizza)))), 1)
        asserte(len(list(triples((Any, likes, cheese)))), 3)
        asserte(len(list(triples((Any, hates, cheese)))), 0)
        asserte(len(list(triples((Any, named, Literal("Bob"))))), 1)

        # unbound objects
        asserte(len(list(triples((michel, likes, Any)))), 2)
        asserte(len(list(triples((tarek, likes, Any)))), 2)
        asserte(len(list(triples((bob, hates, Any)))), 2)
        asserte(len(list(triples((bob, likes, Any)))), 1)
        asserte(len(list(triples((bob, named, Any)))), 1)

        # unbound predicates
        asserte(len(list(triples((michel, Any, cheese)))), 1)
        asserte(len(list(triples((tarek, Any, cheese)))), 1)
        asserte(len(list(triples((bob, Any, pizza)))), 1)
        asserte(len(list(triples((bob, Any, michel)))), 1)
        asserte(len(list(triples((bob, Any, Literal("Bob"))))), 1)

        # unbound subject, objects
        asserte(len(list(triples((Any, hates, Any)))), 2)
        asserte(len(list(triples((Any, likes, Any)))), 5)
        asserte(len(list(triples((Any, named, Any)))), 1)

        # unbound predicates, objects
        asserte(len(list(triples((michel, Any, Any)))), 2)
        asserte(len(list(triples((bob, Any, Any)))), 4)
        asserte(len(list(triples((tarek, Any, Any)))), 2)

        # unbound subjects, predicates
        asserte(len(list(triples((Any, Any, pizza)))), 3)
        asserte(len(list(triples((Any, Any, cheese)))), 3)
        asserte(len(list(triples((Any, Any, michel)))), 1)

        # all unbound
        asserte(len(list(triples((Any, Any, Any)))), 8)
        self.removeStuff(self.graph1)
        asserte(len(list(triples((Any, Any, Any)))), 0)


    def testStatementNode(self):
        graph = self.graph1

        from rdflib.term import Statement
        c = URIRef("http://example.org/foo#c")
        r = Literal("blah")
        s = Statement((self.michel, self.likes, self.pizza), None)
        graph.add((s, RDF.value, r))
        self.assertEqual(len(list(graph.triples((s, RDF.value, r)))), 1)
        self.assertEqual(len(list(graph.triples((s, RDF.value, None)))), 1)
        self.assertEqual(len(list(graph.triples((None, RDF.value, r)))), 1)
        self.assertEqual(len(list(graph.triples((s, None, r)))), 1)
        self.assertEqual(len(list(graph.triples((s, None, None)))), 1)
        self.assertEqual(len(list(graph.triples((None, RDF.value, None)))), 1)
        self.assertEqual(len(list(graph.triples((None, None, r)))), 1)
        self.assertEqual(len(list(graph.triples((None, None, None)))), 2)

        self.assertEquals(r, graph.value(s, RDF.value))
        self.assertEquals(s, graph.value(predicate=RDF.value, object=r))
        self.assertEquals(RDF.value, graph.value(s, None, r))

        graph.remove((s, RDF.value, r))
        self.assertEqual(len(list(graph.triples((s, RDF.value, r)))), 0)

    #def testGraphValue(self):
        #from rdflib.graph import GraphValue

        #graph = self.graph1

        #g1 = Graph(store=self.store_name)
        #g1.open(self.settings1, True)
        #g1.add((self.alice, RDF.value, self.pizza))
        #g1.add((self.bob, RDF.value, self.cheese))
        #g1.add((self.bob, RDF.value, self.pizza))

        #g2 = Graph(store=self.store_name)
        #g2.open(self.settings2, True)
        #g2.add((self.bob, RDF.value, self.pizza))
        #g2.add((self.bob, RDF.value, self.cheese))
        #g2.add((self.alice, RDF.value, self.pizza))

        #gv1 = GraphValue(store=graph.store, graph=g1)
        #gv2 = GraphValue(store=graph.store, graph=g2)
        #graph.add((gv1, RDF.value, gv2))
        #v = graph.value(gv1)
        ##print type(v)
        #self.assertEquals(gv2, v)
        ##print list(gv2)
        ##print gv2.identifier
        #graph.remove((gv1, RDF.value, gv2))

    def testConnected(self):
        graph = self.graph1
        self.addStuff(self.graph1)
        self.assertEquals(True, graph.connected())

        jeroen = self.sNS.jeroen
        unconnected = self.oNS.unconnected

        graph.add((jeroen,self.likes,unconnected))

        self.assertEquals(False, graph.connected())

        # sanity check that we are ignoring reference nodes
        self.assertTrue(graph.store.ignore_reference_nodes)
        # if we don't ignore reference nodes, the graph should be connected
        graph.store.ignore_reference_nodes = False

        self.assertEquals(True, graph.connected())


    def testSub(self):
        g1=self.graph1
        g2=self.graph2

        tarek = self.tarek
        michel = self.michel
        bob = self.bob
        likes = self.likes
        hates = self.hates
        pizza = self.pizza
        cheese = self.cheese
       
        g1.add((tarek, likes, pizza))
        g1.add((bob, likes, cheese))

        g2.add((bob, likes, cheese))

        g3=g1-g2

        self.assertEquals(len(g3), 1)
        self.assertEquals((tarek, likes, pizza) in g3, True)
        self.assertEquals((tarek, likes, cheese) in g3, False)

        self.assertEquals((bob, likes, cheese) in g3, False)

        g1-=g2

        self.assertEquals(len(g1), 1)
        self.assertEquals((tarek, likes, pizza) in g1, True)
        self.assertEquals((tarek, likes, cheese) in g1, False)

        self.assertEquals((bob, likes, cheese) in g1, False)

    def testGraphAdd(self):
        g1=self.graph1
        g2=self.graph2

        tarek = self.tarek
        michel = self.michel
        bob = self.bob
        likes = self.likes
        hates = self.hates
        pizza = self.pizza
        cheese = self.cheese
       
        g1.add((tarek, likes, pizza))

        g2.add((bob, likes, cheese))

        g3=g1+g2

        self.assertEquals(len(g3), 2)
        self.assertEquals((tarek, likes, pizza) in g3, True)
        self.assertEquals((tarek, likes, cheese) in g3, False)

        self.assertEquals((bob, likes, cheese) in g3, True)

        g1+=g2

        self.assertEquals(len(g1), 2)
        self.assertEquals((tarek, likes, pizza) in g1, True)
        self.assertEquals((tarek, likes, cheese) in g1, False)

        self.assertEquals((bob, likes, cheese) in g1, True)

    def testGraphIntersection(self):
        g1=self.graph1
        g2=self.graph2

        tarek = self.tarek
        michel = self.michel
        bob = self.bob
        likes = self.likes
        hates = self.hates
        pizza = self.pizza
        cheese = self.cheese
       
        g1.add((tarek, likes, pizza))
        g1.add((michel, likes, cheese))

        g2.add((bob, likes, cheese))
        g2.add((michel, likes, cheese))

        g3=g1*g2

        self.assertEquals(len(g3), 1)
        self.assertEquals((tarek, likes, pizza) in g3, False)
        self.assertEquals((tarek, likes, cheese) in g3, False)

        self.assertEquals((bob, likes, cheese) in g3, False)

        self.assertEquals((michel, likes, cheese) in g3, True)

        g1*=g2

        self.assertEquals(len(g1), 1)

        self.assertEquals((tarek, likes, pizza) in g1, False)
        self.assertEquals((tarek, likes, cheese) in g1, False)

        self.assertEquals((bob, likes, cheese) in g1, False)

        self.assertEquals((michel, likes, cheese) in g1, True)

    def testSerialize(self):
        node = self.graph1.store.ident_to_node(self.pizza, True)
        log.info("Pizza Attr: %s" %node.attributes)


        parse_serial_pairs = [
            ('xml','xml'),
            # we will add more once we support context aware graphs
        ]
        for parse, serial in parse_serial_pairs:
            self.addStuff(self.graph1)

            v = self.graph1.serialize(format=serial)
            self.graph2.parse(data=v,format=parse)


            for triple in self.graph1:
                self.assertTrue(triple in self.graph2)

            for triple in self.graph2:
                self.assertTrue(triple in self.graph1)

            self.graph1.remove((None,None,None))
            self.graph2.remove((None,None,None))


    def testQuery(self):
        # parse from string
        # borrowed from http://en.wikipedia.org/wiki/Resource_Description_Framework
        rdf = """
            <rdf:RDF
            xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
            xmlns:foaf="http://xmlns.com/foaf/0.1/" 
            xmlns:dc="http://purl.org/dc/elements/1.1/">
                    <rdf:Description rdf:about="http://en.wikipedia.org/wiki/Tony_Benn">
                            <dc:title>Tony Benn</dc:title>
                            <dc:publisher>Wikipedia</dc:publisher>
                            <foaf:primaryTopic>
                                <foaf:Person>
                                    <foaf:name>Tony Benn</foaf:name>  
                                </foaf:Person>
                            </foaf:primaryTopic>
                    </rdf:Description>
            </rdf:RDF>
        """

        self.graph1.parse(data=rdf)
        rdflib.plugin.register('sparql', rdflib.query.Processor,
                        'rdfextras.sparql.processor', 'Processor')
        rdflib.plugin.register('sparql', rdflib.query.Result,
                        'rdfextras.sparql.query', 'SPARQLQueryResult')
        rows = self.graph1.query(
            """
            SELECT ?a WHERE 
            { ?a foaf:primaryTopic ?b . ?b foaf:name "Tony Benn" }
            """,
            initNs=dict(self.graph1.namespaces())
        )

        self.assertEqual(len(rows), 1)

    def testParse(self):
        # examples from w3c

        # TODO: this fails because of query string in url
        #self.graph1.parse("http://www.w3.org/2000/10/rdf-tests/rdfcore/amp-in-url/test001.rdf")
        self.graph1.parse("http://www.w3.org/2000/10/rdf-tests/rdfcore/datatypes/test001.rdf")
        # TODO: this fails due to type parsing, probably an rdflib problem
        #self.graph1.parse("http://www.w3.org/2000/10/rdf-tests/rdfcore/datatypes/test002.rdf")

        self.graph1.parse("http://www.w3.org/2000/10/rdf-tests/rdfcore/rdf-element-not-mandatory/test001.rdf")
        self.graph1.parse("http://www.w3.org/2000/10/rdf-tests/rdfcore/rdfms-reification-required/test001.rdf")
        self.graph1.parse("http://www.w3.org/2000/10/rdf-tests/rdfcore/rdfms-uri-substructure/test001.rdf")

        self.graph1.parse("http://www.w3.org/2000/10/rdf-tests/rdfcore/rdfms-xmllang/test001.rdf")
        self.graph1.parse("http://www.w3.org/2000/10/rdf-tests/rdfcore/rdfms-xmllang/test002.rdf")
        self.graph1.parse("http://www.w3.org/2000/10/rdf-tests/rdfcore/rdfms-xmllang/test003.rdf")

        self.graph1.parse("http://www.w3.org/2000/10/rdf-tests/rdfcore/rdfms-xmllang/test004.rdf")
        self.graph1.parse("http://www.w3.org/2000/10/rdf-tests/rdfcore/rdfms-xmllang/test005.rdf")
        self.graph1.parse("http://www.w3.org/2000/10/rdf-tests/rdfcore/rdfms-xmllang/test006.rdf")

        self.graph1.parse("http://www.w3.org/2000/10/rdf-tests/rdfcore/unrecognised-xml-attributes/test001.rdf")
        self.graph1.parse("http://www.w3.org/2000/10/rdf-tests/rdfcore/unrecognised-xml-attributes/test002.rdf")
        self.graph1.parse("http://www.w3.org/2000/10/rdf-tests/rdfcore/xml-canon/test001.rdf")

        #additional examples for the fun of it
        self.graph1.parse("http://bigasterisk.com/foaf.rdf")
        self.graph1.parse("http://www.w3.org/People/Berners-Lee/card.rdf")
        self.graph1.parse("http://danbri.livejournal.com/data/foaf") 

        self.graph1.serialize("serialized.rdf")


class GraphCassandraTestCase(GraphMemoryTestCase):
    settings1 = {
        'agamemnon.keyspace' : 'testagamemnon1',
        'agamemnon.host_list' : '["localhost:9160"]',
        'agamemnon.rdf_namespace_base' : 'http://www.example.org/rdf/',
    }
    settings2 = {
        'agamemnon.keyspace' : 'testagamemnon2',
        'agamemnon.host_list' : '["localhost:9160"]',
        'agamemnon.rdf_namespace_base' : 'http://www.example.org/rdf/',
    }


if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase(GraphMemoryTestCase)
    unittest.TextTestRunner(verbosity=2).run(suite)





from rdflib.store import Store, NO_STORE, VALID_STORE
from rdflib.plugin import register
from rdflib.term import URIRef, Literal, BNode, Statement
from rdflib.namespace import Namespace, split_uri
from agamemnon.factory import load_from_settings
from agamemnon.exceptions import NodeNotFoundException
from pycassa.cassandra.ttypes import NotFoundException
from urlparse import urlsplit, urlunsplit, urldefrag
import pycassa
import json
import uuid
import logging

register('Agamemnon', Store, 
                'agamemnon.rdf_store', 'AgamemnonStore')

log = logging.getLogger(__name__)

RDF_NAMESPACE_CF = "__RDF_NAMESPACE_CF"
RDF_DEFAULT_PREFIX = "__DEFAULT__PREFIX"

BNODE_NODE_TYPE = "__BNODE_NODE_TYPE"

class AgamemnonStore(Store):
    """
    An agamemnon based triple store.
    
    This triple store uses agamemnon as the underlying graph representation.
    
    """

    
    def __init__(self, configuration=None, identifier=None, data_store=None):
        super(AgamemnonStore, self).__init__(configuration)
        self.identifier = identifier

        # namespace and prefix indexes
        self._ignored_node_types = set(['reference'])

        self.configuration = configuration or dict()

        if data_store:
            self.data_store = data_store



    def open(self, configuration=None, create=False, repl_factor = 1):
        if configuration:
            self.configuration = configuration
        keyspace = self.configuration['agamemnon.keyspace']
        if create and keyspace != "memory":
            hostlist = json.loads(self.configuration['agamemnon.host_list'])
            system_manager = pycassa.SystemManager(hostlist[0])
            try:
                log.info("Attempting to drop keyspace: %s" % keyspace)
                system_manager.drop_keyspace(keyspace)
            except pycassa.cassandra.ttypes.InvalidRequestException:
                log.warn("Keyspace didn't exist")
            finally:
                log.info("Creating keyspace: %s" % keyspace)
                system_manager.create_keyspace(keyspace, replication_factor=repl_factor)

        self.data_store = load_from_settings(self.configuration)

        return VALID_STORE

    @property
    def data_store(self):
        return self._ds

    @data_store.setter
    def data_store(self, ds):
        self._ds = ds
        self._process_config()

    @property
    def ignore_reference_nodes(self):
        return "reference" in self._ignored_node_types

    @ignore_reference_nodes.setter
    def ignore_reference_nodes(self, value):
        if value:
            self.ignore('reference')
        else:
            self.unignore('reference')

    @property
    def namespace_base(self):
        return self.namespace("")

    @namespace_base.setter
    def namespace_base(self, value):
        self.bind("",value)

    def ignore(self, node_type):
        self._ignored_node_types.add(node_type)

    def unignore(self, node_type):
        self._ignored_node_types.remove(node_type)

    def _process_config(self):
        config_prefix = "agamemnon.rdf_"
        for key, value in self.configuration.items():
            if key.startswith(config_prefix):
                setattr(self, key[len(config_prefix):], value)

    def add(self, (subject, predicate, object), context, quoted=False):
        log.debug("Adding  %r, %r, %r" % (subject, predicate, object))
        if isinstance(subject, Literal):
            raise TypeError("Subject can't be literal")

        if isinstance(predicate, Literal):
            raise TypeError("Predicate can't be literal")

        if isinstance(predicate, Statement):
            #TODO: support sentence objects
            raise TypeError("Agamemnon doesn't support sentential objects.")

        p_rel_type = self.ident_to_rel_type(predicate) 

        #inline literals as attributes
        if isinstance(subject, Statement):
            if not isinstance(object, Literal):
                raise TypeError("Agamemnon requires sentence objects to be literal.")
            for rel in self._statement_to_rel(subject, create=True):
                # TODO: this loop should execute only once ... should we sanity check it?
                rel[p_rel_type] = object.toPython()
                rel.commit()
        elif isinstance(object, Literal):
            s_node = self.ident_to_node(subject, True)
            log.debug("Setting %r on %s" % (p_rel_type, s_node))
            s_node[p_rel_type] = object.toPython()
            s_node.commit()
        else:
            s_node = self.ident_to_node(subject, True)
            o_node = self.ident_to_node(object, True)

            log.debug("Creating relationship of type %s from %s on %s" % (p_rel_type, s_node, o_node))
            self.data_store.create_relationship(str(p_rel_type), s_node, o_node)

    def remove(self, triple, context=None):
        for (subject, predicate, object), c in self.triples(triple):
            log.debug("Deleting %s, %s, %s" % (subject, predicate, object))
            p_rel_type = self.ident_to_rel_type(predicate)
            if isinstance(subject, Statement):
                for rel in self._statement_to_rel(subject):
                    del rel[p_rel_type]
                    rel.commit()
            else:
                s_node = self.ident_to_node(subject)
                if isinstance(object, Literal):
                    if p_rel_type in s_node.attributes:
                        if s_node[p_rel_type] == object.toPython():
                            del s_node[p_rel_type]
                            s_node.commit()

                            s_node = self.data_store.get_node(s_node.type, s_node.key)
                            log.debug("ATTS: %s" %s_node.attributes)

                else:
                    o_node_type, o_node_id = self.ident_to_node_def(object) 
                    if o_node_type in self._ignored_node_types: return
                    for rel in getattr(s_node, p_rel_type).relationships_with(o_node_id):
                        if rel.target_node.type == o_node_type:
                            rel.delete()

    def triples(self, (subject, predicate, object), context=None):
        log.debug("Looking for triple %s, %s, %s" % (subject, predicate, object))
        if isinstance(subject, Literal) or isinstance(predicate, Literal):
            # subject and predicate can't be literal silly rabbit
            return 

        # Determine what mechanism to use to do lookup
        try:
            if predicate is not None:
                if subject is not None:
                    if object is not None:
                        triples = self._triples_by_spo(subject, predicate, object)
                    else:
                        triples = self._triples_by_sp(subject, predicate)
                else:
                    if object is not None:
                        triples = self._triples_by_po(predicate, object)
                    else:
                        triples = self._triples_by_p(predicate)
            else:
                if subject is not None:
                    if object is not None:
                        triples = self._triples_by_so(subject, object)
                    else:
                        triples = self._triples_by_s(subject)
                else:
                    if object is not None:
                        triples = self._triples_by_o(object)
                    else:
                        triples = self._all_triples()

            for triple in triples:
                yield triple, None
        except NodeNotFoundException:
            # exit generator as we found no triples
            log.debug("Failed to find any triples.")
            return

    def _triples_by_spo(self, subject, predicate, object):
        log.debug("Finding triple by spo")
        if isinstance(subject, Statement):
            for rel in self._statement_to_rel(subject):
                for triple in self._relationship_triples(rel,search_pred=predicate,search_obj=object):
                    yield triple
        else:
            p_rel_type = self.ident_to_rel_type(predicate) 
            s_node = self.ident_to_node(subject)
            if s_node.type in self._ignored_node_types: return
            if isinstance(object, Literal):
                if p_rel_type in s_node.attributes:
                    if s_node[p_rel_type] == object.toPython():
                        log.debug("Found %s, %s, %s" % (subject, predicate, object))
                        yield subject, predicate, object
            else:
                o_node_type, o_node_id = self.ident_to_node_def(object) 
                if o_node_type in self._ignored_node_types: return
                for rel in getattr(s_node, p_rel_type).relationships_with(o_node_id):
                    if rel.target_node.type == o_node_type:
                        log.debug("Found %s, %s, %s" % (subject, predicate, object))
                        yield subject, predicate, object

    def _triples_by_sp(self, subject, predicate):
        log.debug("Finding triple by sp")
        if isinstance(subject, Statement):
            for rel in self._statement_to_rel(subject):
                for triple in self._relationship_triples(rel,search_pred=predicate):
                    yield triple
        else:
            p_rel_type = self.ident_to_rel_type(predicate) 
            s_node = self.ident_to_node(subject) 
            if s_node.type in self._ignored_node_types: return
            for rel in getattr(s_node, p_rel_type).outgoing:
                object = self.node_to_ident(rel.target_node)
                log.debug("Found %s, %s, %s" % (subject, predicate, object))
                yield subject, predicate, object

            if p_rel_type in s_node.attributes:
                object = Literal(s_node[p_rel_type])
                log.debug("Found %s, %s, %s" % (subject, predicate, object))
                yield subject, predicate, object

    def _triples_by_po(self, predicate, object):
        log.debug("Finding triple by po")
        p_rel_type = self.ident_to_rel_type(predicate) 
        if isinstance(object, Literal):
            log.warn("Your query requires full graph traversal do to Agamemnon datastructure.")
            for s_node in self._all_nodes():
                subject = self.node_to_ident(s_node)
                if p_rel_type in s_node.attributes:
                    if s_node[p_rel_type] == object.toPython():
                        log.debug("Found %s, %s, %s" % (subject, predicate, object))
                        yield subject, predicate, object
                for rel in s_node.relationships.outgoing:
                    for triple in self._relationship_triples(rel,search_obj=object, search_pred=predicate):
                        yield triple
        else:
            o_node = self.ident_to_node(object) 
            for rel in getattr(o_node, p_rel_type).incoming:
                subject = self.node_to_ident(rel.source_node)
                log.debug("Found %s, %s, %s" % (subject, predicate, object))
                yield subject, predicate, object

    def _triples_by_so(self, subject, object):
        log.debug("Finding triple by so.")
        if isinstance(subject, Statement):
            for rel in self._statement_to_rel(subject):
                for triple in self._relationship_triples(rel,search_obj=object):
                    yield triple
        else:
            s_node = self.ident_to_node(subject) 
            if s_node.type in self._ignored_node_types: return
            if isinstance(object, Literal):
                for p_rel_type in s_node.attributes.keys():
                    if p_rel_type.startswith("__"): continue #ignore special names
                    if s_node[p_rel_type] == object.toPython():
                        predicate = self.rel_type_to_ident(p_rel_type)
                        log.debug("Found %s, %s, %r" % (subject, predicate, object))
                        yield subject, predicate, object
            else:
                o_node = self.ident_to_node(object) 
                if o_node.type in self._ignored_node_types: return
                for rel in s_node.relationships.outgoing:
                    if rel.target_node == o_node:
                        predicate = self.rel_type_to_ident(rel.type)
                        log.debug("Found %s, %s, %s" % (subject, predicate, object))
                        yield subject, predicate, object

    def _triples_by_s(self, subject):
        log.debug("Finding triple by s")
        if isinstance(subject, Statement):
            for rel in self._statement_to_rel(subject):
                for triple in self._relationship_triples(rel):
                    yield triple
        else:
            s_node = self.ident_to_node(subject) 
            if s_node.type in self._ignored_node_types: return
            for rel in s_node.relationships.outgoing:
                if rel.target_node.type in self._ignored_node_types: continue
                predicate = self.rel_type_to_ident(rel.type)
                object = self.node_to_ident(rel.target_node)
                log.debug("Found %s, %s, %s" % (subject, predicate, object))
                yield subject, predicate, object

            for p_rel_type in s_node.attributes.keys():
                if p_rel_type.startswith("__"): continue #ignore special names
                predicate = self.rel_type_to_ident(p_rel_type)
                object = Literal(s_node[p_rel_type])
                log.debug("Found %s, %s, %r" % (subject, predicate, object))
                yield subject, predicate, object

    def _triples_by_p(self, predicate):
        log.debug("Finding triple by p")
        log.warn("Your query requires full graph traversal do to Agamemnon datastructure.")

        p_rel_type = self.ident_to_rel_type(predicate) 
        for s_node in self._all_nodes():
            if s_node.type in self._ignored_node_types: continue
            subject = self.node_to_ident(s_node)
            for rel in s_node.relationships.outgoing:
                for triple in self._relationship_triples(rel,search_pred=predicate):
                    yield triple
                if rel.type != p_rel_type: continue
                if rel.target_node.type in self._ignored_node_types: continue
                object = self.node_to_ident(rel.target_node)
                log.debug("Found %s, %s, %s" % (subject, predicate, object))
                yield subject, predicate, object


            if p_rel_type in s_node.attributes:
                object = Literal(s_node[p_rel_type])
                log.debug("Found %s, %s, %s" % (subject, predicate, object))
                yield subject, predicate, object 

    def _triples_by_o(self, object):
        log.debug("Finding triple by o")
        if isinstance(object, Literal):
            log.warn("Your query requires full graph traversal do to Agamemnon datastructure.")
            for s_node in self._all_nodes():
                if s_node.type in self._ignored_node_types: continue
                subject = self.node_to_ident(s_node)
                for p_rel_type in s_node.attributes.keys():
                    if p_rel_type.startswith("__"): continue #ignore special names
                    if s_node[p_rel_type] == object.toPython():
                        predicate = self.rel_type_to_ident(p_rel_type)
                        log.debug("Found %s, %s, %s" % (subject, predicate, object))
                        yield subject, predicate, object
                for rel in s_node.relationships.outgoing:
                    for triple in self._relationship_triples(rel):
                        yield triple

        else:
            o_node = self.ident_to_node(object) 
            for rel in o_node.relationships.incoming:
                if rel.source_node.type in self._ignored_node_types: continue
                predicate = self.rel_type_to_ident(rel.type)
                subject = self.node_to_ident(rel.source_node)
                log.debug("Found %s, %s, %s" % (subject, predicate, object))
                yield subject, predicate, object

    def _all_triples(self):
        log.debug("Finding all triples.")
        log.warn("Your query requires full graph traversal do to Agamemnon datastructure.")
        for s_node in self._all_nodes():
            if s_node.type in self._ignored_node_types: continue
            subject = self.node_to_ident(s_node)
            for rel in s_node.relationships.outgoing:
                if rel.target_node.type in self._ignored_node_types: continue
                predicate = self.rel_type_to_ident(rel.type)
                object = self.node_to_ident(rel.target_node)
                log.debug("Found %s, %s, %s" % (subject, predicate, object))
                yield subject, predicate, object

                for triple in self._relationship_triples(rel):
                    yield triple

            for p_rel_type in s_node.attributes.keys():
                if p_rel_type.startswith("__"): continue #ignore special names
                predicate = self.rel_type_to_ident(p_rel_type)
                object = Literal(s_node[p_rel_type])
                log.debug("Found %s, %s, %s" % (subject, predicate, object))
                yield subject, predicate, object

    def _all_nodes(self):
        ref_ref_node = self.data_store.get_reference_node()
        for ref in ref_ref_node.instance.outgoing:
            if ref.target_node.key in self._ignored_node_types: continue
            for instance in ref.target_node.instance.outgoing:
                yield instance.target_node

    def _relationship_triples(self, rel, search_pred=None, search_obj=None):
        rel_subject = self.rel_to_statement(rel)
        for p_rel_type in rel.attributes.keys():
            if p_rel_type.startswith("__"): continue #ignore special names
            object = Literal(rel[p_rel_type])
            predicate = self.rel_type_to_ident(p_rel_type)
            if search_pred and search_pred != predicate: continue
            if search_obj and search_obj != object: continue
            log.debug("Found %s, %s, %s" % (rel_subject, predicate, object))
            yield rel_subject, predicate, object

    def _statement_to_rel(self, statement, create=False):
        (subject, predicate, object), _ = statement
        if isinstance(object, Literal):
            raise TypeError("Statements with literal objects not supported.")

        s_node = self.ident_to_node(subject, create=True)
        p_rel_type = self.ident_to_rel_type(predicate)
        o_node_type, o_node_key = self.ident_to_node_def(object)
        for rel in getattr(s_node, p_rel_type).relationships_with(o_node_key):
            if rel.target_node.type == o_node_type:
                yield rel
                return

        if create:
            o_node = self.ident_to_node(object, create=True)
            rel = getattr(s_node,p_rel_type)(o_node)
            yield rel


    def rel_to_statement(self, rel):
        subject = self.node_to_ident(rel.source_node)
        predicate = self.rel_type_to_ident(rel.type)
        object = self.node_to_ident(rel.target_node)
        statement = Statement((subject, predicate, object), None)
        log.debug("Converted relationship %s to: %s" % (rel, statement))
        return statement

    def node_to_ident(self, node):
        if node.type == BNODE_NODE_TYPE:
            return BNode(node.key)
        else:
            ns = self.namespace(node.type)
            if ns is None:
                ns = Namespace(self.namespace_base[node.type + "#"])
                self.bind(node.type, ns)
            uri = ns[node.key]
            log.debug("Converted node %s to uri %s" % (node, uri))
            return uri

    def ident_to_node(self, identifier, create=False):
        node_type, node_id = self.ident_to_node_def(identifier)

        try:
            log.debug("Looking up node: %s => %s" % (node_type,node_id))
            return self.data_store.get_node(node_type, node_id)
        except NodeNotFoundException:
            if create:
                node = self.data_store.create_node(node_type, node_id)
                log.debug("Created node: %s" % node)
            else:
                raise
            return node

    def ident_to_node_def(self, identifier):
        if isinstance(identifier,URIRef): 
            namespace, node_id = self._split_uri(identifier)

            node_type = self.prefix(namespace)
            if node_type is None:
                node_type = str(uuid.uuid1()).replace("-","_")
                self.bind(node_type, namespace)
            elif node_type == "":
                node_type = RDF_DEFAULT_PREFIX

            return node_type, node_id
        elif isinstance(identifier,BNode): 
            # Bnodes get their own table
            node_type = BNODE_NODE_TYPE
            node_id = identifier.encode("utf-8")
            return node_type, node_id
        else:
            raise ValueError("Unknown identifier type %r" % identifier.__class__)

    def rel_type_to_ident(self, rel_type):
        if ":" in rel_type:
            prefix, suffix = rel_type.split(":",1)
            namespace = self.namespace(prefix)
            if namespace:
                identifier = namespace[suffix]
            else:
                identifier = URIRef(rel_type)
        else:
            identifier = self.namespace("")[rel_type]

        return identifier

    def ident_to_rel_type(self, identifier):
        namespace, rel_type = self._split_uri(identifier)
        prefix = self.prefix(namespace)
        if prefix is None:
            prefix = str(uuid.uuid1()).replace("-","_")
            self.bind(prefix, rel_type)
        
        if prefix != "":
            rel_type = ":".join((prefix,rel_type))
        
        return rel_type.encode('utf-8')

    def _split_uri(self, identifier):
        if isinstance(identifier,URIRef): 
            scheme, netloc, path, query, fragment = urlsplit(identifier)
            if query:
                namespace, resource_id = split_uri(identifier)
            if fragment:
                # if we have a fragment, we will split there
                namespace, resource_id = urldefrag(identifier)
                namespace += "#"
            elif "/" in path and len(path)>1:
                splits = path.split("/")
                if path.endswith("/"):
                    resource_id = "/".join(splits[-2:])
                    path = "/".join(splits[:-2]) + "/"
                    namespace = urlunsplit((scheme, netloc, path, "", ""))
                else:
                    resource_id = "/".join(splits[-1:])
                    path = "/".join(splits[:-1]) + "/"
                    namespace = urlunsplit((scheme, netloc, path, "", ""))
            elif path:
                resource_id = path
                namespace = urlunsplit((scheme, netloc, "", "", ""))
            else:
                namespace, resource_id = split_uri(identifier)

            log.debug("Split %s to %s, %s" % (identifier, namespace, resource_id))
            return namespace, resource_id
        else:
            raise ValueError("Unknown identifier type %r" % identifier)

    def bind(self, prefix, namespace):
        log.debug("Binding prefix '%s' to namespace %s" % (prefix, namespace))
        if not prefix:
            prefix = RDF_DEFAULT_PREFIX
        prefix = prefix.encode("utf-8")
        namespace = namespace.encode("utf-8")
        self.data_store.insert(RDF_NAMESPACE_CF, 'namespace', {prefix:namespace})
        self.data_store.insert(RDF_NAMESPACE_CF, 'prefix', {namespace:prefix})


    def namespace(self, prefix):
        log.debug("Finding namespace for %s" % prefix)
        if not prefix:
            prefix = RDF_DEFAULT_PREFIX
        try:
            prefix = prefix.encode("utf-8")
            entry = self.data_store.get(
                RDF_NAMESPACE_CF, 'namespace', 
                column_start=prefix, column_finish=prefix, column_count = 1
            )
            if prefix in entry:
                return Namespace(entry[prefix])
            else:
                return None
        except NotFoundException:
            return None

    def prefix(self, namespace):
        log.debug("Finding prefix for %s" % namespace)
        try:
            namespace = namespace.encode("utf-8")
            entry = self.data_store.get(
                RDF_NAMESPACE_CF, 'prefix', 
                column_start=namespace, column_finish=namespace, column_count = 1
            )
            prefix = entry.get(namespace,None)
            if prefix == RDF_DEFAULT_PREFIX:
                prefix = ""
            return prefix
        except NotFoundException:
            return None

    def namespaces(self):
        try:
            entry = self.data_store.get(RDF_NAMESPACE_CF, 'namespace', column_count=1000)
            for prefix, namespace in entry.items():
                if prefix == RDF_DEFAULT_PREFIX:
                    prefix = ""
                yield prefix, Namespace(namespace)
        except NotFoundException:
            return 

    def __contexts(self):
        return (c for c in []) # TODO: best way to return empty generator

    def __len__(self, context=None):
        return len(list(self._all_triples()))




import json
from pycassa import system_manager
import pycassa
from pycassa.batch import Mutator
from pycassa.cassandra.ttypes import NotFoundException
from agamemnon.graph_constants import OUTBOUND_RELATIONSHIP_CF, INBOUND_RELATIONSHIP_CF, RELATIONSHIP_INDEX
import pycassa.columnfamily as cf

class CassandraDataStore(object):
    def __init__(self, keyspace, pool, system_manager):
        self._cf_cache = {}
        self._index_cache = {}
        self._system_manager = system_manager
        self._pool = pool
        self._keyspace = keyspace
        self._batch = None
        self.in_batch = False
        self.batch_count = 0
        if not self.cf_exists(OUTBOUND_RELATIONSHIP_CF):
            self.create_cf(OUTBOUND_RELATIONSHIP_CF, super=True)
        if not self.cf_exists(INBOUND_RELATIONSHIP_CF):
            self.create_cf(INBOUND_RELATIONSHIP_CF, super=True)
        if not self.cf_exists(RELATIONSHIP_INDEX):
            self.create_cf(RELATIONSHIP_INDEX, super=True)

    def create_cf(self, type, column_type=system_manager.ASCII_TYPE, super=False, index_columns=list()):
        self._system_manager.create_column_family(self._keyspace, type, super=super, comparator_type=column_type)
        for column in index_columns:
            self._system_manager.create_index(self._keyspace, type, column, column_type,
                                              index_name='%s_%s_index' % (type, column))
        return cf.ColumnFamily(self._pool, type)
    
    def cf_exists(self, type):
        try:
            cf.ColumnFamily(self._pool, type)
        except NotFoundException:
            return False
        return True

    def get_cf(self, type, create=True):

        column_family = None
        try:
            column_family = cf.ColumnFamily(self._pool, type)
            self._cf_cache[type] = column_family
        except NotFoundException:
            if create:
                column_family = self.create_cf(type)
        return column_family



    def insert(self, column_family, key, columns):
        if self._batch is not None:
            self._batch.insert(column_family, key, columns)
        with Mutator(self._pool) as b:
            b.insert(column_family, key, columns)

    def remove(self,column_family, key, columns=None, super_column=None):
        if self._batch is not None:
            self._batch.remove(column_family, key, columns=columns, super_column=super_column)
        else:
            column_family.remove(key, columns=columns, super_column=super_column)

    def start_batch(self):
        if self._batch is None:
            self.in_batch = True
            self._batch = Mutator(self._pool)
        self.batch_count += 1


    def commit_batch(self):
        self.batch_count -= 1
        if not self.batch_count:
            self._batch.send()
            self._batch = None

def drop_keyspace(host_list, keyspace):
    system_manager = pycassa.SystemManager(json.loads(host_list)[0])
    system_manager.drop_keyspace(keyspace)

def create_keyspace(host_list, keyspace, replication_factor=1):
    system_manager = pycassa.SystemManager(json.loads(host_list)[0])
    system_manager.create_keyspace(keyspace, replication_factor=replication_factor)
